# pyright: reportUnusedFunction=false
from typing import TYPE_CHECKING, cast

import jinja2
import markdown
from jinja2.utils import markupsafe  # type: ignore
from klein import Klein
from klein.resource import KleinResource
from twisted.logger import Logger
from twisted.python.filepath import FilePath
from twisted.web import resource, static
from twisted.web.server import Request

from adlermanager.AdlerManagerTokenResource import AdlerManagerTokenResource
from adlermanager.Config import Config

if TYPE_CHECKING:
    from adlermanager.SitesManager import SiteManager, SitesManager

log = Logger()


def get_jinja_env(support_dir: str) -> jinja2.Environment:
    """
    Return a L{jinja2.Environment} with templates loaded from:
      - Package
      - Support dir

    @param support_dir: Full path to support dir.
      See L{authapiv02.DefaultConfig.Config}
    @type support_dir: L{str}
    """
    md = markdown.Markdown(
        extensions=[
            "markdown.extensions.toc",
            "markdown.extensions.tables",
        ]
    )
    templates = jinja2.Environment(
        extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
        loader=jinja2.ChoiceLoader(
            [
                jinja2.FileSystemLoader(support_dir),
                jinja2.PackageLoader("adlermanager", "templates"),
            ]
        ),
        autoescape=True,
    )

    def md_filter(txt: str) -> markupsafe.Markup:
        return markupsafe.Markup(md.convert(txt))

    templates.filters["markdown"] = md_filter  # type: ignore
    return templates


def web_root(sites_manager: "SitesManager") -> KleinResource:
    app = Klein()

    @app.route("/")
    def index(request: Request):  # type: ignore
        try:
            host = cast(str, request.getRequestHostname().decode("utf-8"))
        except Exception:
            return resource.ErrorPage(400, "Bad cat", '<a href="http://http.cat/400">http://http.cat/400</a>')
        if host not in sites_manager.site_managers:
            return resource.ErrorPage(404, "Gone cat", '<a href="http://http.cat/404">http://http.cat/404</a>')
        site: SiteManager
        try:
            site = sites_manager.site_managers[host]
        except Exception:
            log.failure("sad cat")
            return resource.ErrorPage(500, "Sad cat", '<a href="http://http.cat/500">http://http.cat/500</a>')

        site_path = cast(  # type: ignore
            str, FilePath(Config.data_dir).child("sites").child(host).path
        )
        templates = get_jinja_env(site_path)
        template = templates.get_template("template.j2")

        return template.render(site=site)

    @app.route("/api/v2/alerts", methods=["POST"])
    def alert_handler(request: Request):  # type: ignore  # noqa: ARG001
        return AdlerManagerTokenResource(sites_manager)

    @app.route("/static", branch=True)
    def static_files(request: Request):  # type: ignore  # noqa: ARG001
        return static.File(Config.web_static_dir)

    @app.route("/vendor", branch=True)
    def vendor_files(request: Request):  # type: ignore  # noqa: ARG001
        return static.File(FilePath(__file__).parent().child("vendor").path)

    return app.resource()



import json
from typing import TYPE_CHECKING, Any, Dict, List

from twisted.internet import reactor, task
from twisted.web._responses import BAD_REQUEST, OK
from twisted.web.server import Request

from adlermanager.TokenResource import TokenResource

if TYPE_CHECKING:
    from adlermanager.SitesManager import SiteManager, SitesManager


class AdlerManagerTokenResource(TokenResource):
    """
    TokenResource used for AdlerManager
    """

    HEADER = "Authorization"

    def __init__(self, sites_manager: "SitesManager"):
        """
        @param site_manager: The object managing state for all sites.
        @type  site_manager: L{adlermanager.SitesManager}
        """
        TokenResource.__init__(self, tokens=sites_manager.tokens)

    def preprocess_header(self, header: str) -> str:
        return header.split(" ")[-1]

    def processToken(self, token_data: "SiteManager", request: Request) -> int:
        """
        Pass Alerts along if Authorization Header matched.

        @param token_data: The object associated with the passed token.
        @type  L{adlermanager.SiteManager}

        @param request: The request object associated to this request.
        @type  L{twisted.web.http.Request}
        """

        try:
            request_body: bytes = request.content.read()  # type: ignore
            alert_data: List[Dict[str, Any]] = json.loads(request_body)
        except Exception:
            return BAD_REQUEST

        site = token_data

        task.deferLater(reactor, 0, site.process_alerts, alert_data)  # type: ignore
        return OK



from typing import Dict, Iterable, List, Optional

import attr
from twisted.internet import defer, reactor, task
from twisted.python.filepath import FilePath

from adlermanager.Config import ConfigClass
from adlermanager.model import Alert, Severity
from adlermanager.utils import current_timestamp, default_errback, noop_deferred

FILENAME_TIME_FORMAT = "%Y-%m-%d-%H%MZ"


@attr.s
class IncidentManager:
    global_config: ConfigClass = attr.ib()
    path: FilePath = attr.ib()
    timestamp: str = attr.ib(default="")

    last_alert: str = attr.ib(default="")
    active_alerts: Dict[str, Alert] = attr.ib(factory=dict)
    expired: defer.Deferred[None] = attr.ib(factory=noop_deferred)
    _timeout: defer.Deferred[None] = attr.ib(factory=noop_deferred)
    """Incident timeout"""
    _alert_timeouts: Dict[str, defer.Deferred[None]] = attr.ib(factory=dict)
    """alert_label -> timeout"""

    _monitoring_down: bool = attr.ib(default=False)

    @property
    def incident_grouping_seconds(self) -> float:
        return self.global_config.group_incidents_minutes.total_seconds()

    @property
    def alert_resolve_seconds(self) -> float:
        return self.global_config.alert_resolve_minutes.total_seconds()

    def __attrs_post_init__(self) -> None:
        if not self.path.isdir():
            self.path.createDirectory()

        timestamp_file = self.path.child("timestamp")

        if not self.timestamp:
            if timestamp_file.isfile():
                # The file exists already, read it
                self.timestamp = timestamp_file.getContent().decode("utf-8")
            else:
                # Get timestamp implicitly from path, this is likely a new incident
                #
                # We don't always rely on this, to allow for incident
                # re-naming later on.
                self.timestamp = self.path.basename()

        if not timestamp_file.isfile():
            # Persist the timestamp file if necessary
            timestamp_file.setContent(self.timestamp.encode("utf-8"))
            timestamp_file.chmod(0o644)

    def process_heartbeats(self, heartbeats: Iterable[Alert], timestamp: str) -> None:
        if heartbeats:
            self.last_updated = timestamp
            if self._monitoring_down:
                self._monitoring_down = False
                # Monitoring is back up, re-activate timeout
                self._timeout = task.deferLater(
                    reactor,  # type: ignore
                    self.incident_grouping_seconds,
                    self._expire,
                )
                self.log_event("[Meta]MonitoringUp", timestamp)

    def process_alerts(self, alerts: Iterable[Alert], timestamp: str) -> None:
        if alerts:
            self._timeout.cancel()
            self._timeout = task.deferLater(
                reactor,  # type: ignore
                self.incident_grouping_seconds,
                self._expire,
            ).addErrback(default_errback)
            self.last_alert = timestamp

        new_alerts: Dict[str, Alert] = {}

        for alert in alerts:
            alert_label = alert.labels["component"]
            if alert_label in self._alert_timeouts:
                self._alert_timeouts[alert_label].cancel()
            else:
                new_alerts[alert_label] = alert
            # Use highest known severity for a given incident
            # We also allow Resolved alert notifications
            if alert.status == Severity.OK or alert.status >= self.active_alerts.get(alert_label, alert).status:
                self.active_alerts[alert_label] = alert
            self._alert_timeouts[alert_label] = task.deferLater(
                reactor,  # type: ignore
                self.alert_resolve_seconds,
                self._expire_alert,
                alert_label,
            ).addErrback(default_errback)

        if new_alerts:
            self.log_event("New", timestamp, alerts=list(new_alerts.values()))

    def _expire(self) -> None:
        if not self._monitoring_down:
            self.expired.callback(self)  # type: ignore  # twisted bug

    def _expire_alert(self, alert_label: str) -> None:
        self.log_event("Resolved", current_timestamp(), alert=self.active_alerts[alert_label])
        del self.active_alerts[alert_label]

    def monitoring_down(self, timestamp: str) -> None:
        self._monitoring_down = True
        self.log_event("[Meta]MonitoringDown", timestamp)

    def log_event(
        self,
        message: str,
        timestamp: str,
        alerts: Optional[List[Alert]] = None,
        alert: Optional[Alert] = None,
    ) -> None:
        # TODO: we probably actually want something like this
        pass

    # def log_event(
    #     self,
    #     message: str,
    #     timestamp: str,
    #     alerts: List[Alert] = [],
    #     alert: Optional[Alert] = None,
    # ):
    #     if alert is not None:
    #         alerts.append(alert)
    #     obj = {"message": message, "timestamp": timestamp}
    #     if alerts:
    #         alerts = copy.deepcopy(alerts)
    #         for a in alerts:
    #             a.status = a.status.css
    #         obj["alerts"] = alerts
    #     # self._logs.append(obj)
    #     # Persist messages
    #     with self.path.open("w") as f:
    #         m = Munch.fromDict({"log": self._logs, "timestamp": timestamp})
    #         f.write(m.toYAML().encode("utf-8"))

    def component_status(self, component_label: str) -> Severity:
        return max(
            (alert.status for alert in self.active_alerts.values() if alert.labels["component"] == component_label),
            default=Severity.OK,
        )



from typing import Any, Dict, Generator, Optional, Union, cast

from twisted.internet import defer
from twisted.logger import Logger
from twisted.web import resource, server
from twisted.web._responses import INTERNAL_SERVER_ERROR, OK, UNAUTHORIZED
from twisted.web.server import Request

log = Logger()


class TokenResource(resource.Resource):
    """
    A simple Header based Auth-Token protected Resource.

    Subclass and replace HEADER, processToken and unauthorizedMessage
    as needed.

    @cvar HEADER: The header to get the token from.
    @ivar tokens: A mapping of valid tokens to any target objects.
    """

    HEADER = "Auth-Token"

    def __init__(self, tokens: Optional[Dict[str, Any]] = None):
        """
        @param tokens: The mapping of valid tokens to target objects.
        """
        resource.Resource.__init__(self)
        self.tokens = {} if tokens is None else tokens

    def render(self, request: Request) -> Union[bytes, int]:
        """
        See L{resource.Resource}.

        @see: L{resource.Resource.render}.
        """
        raw_header = request.getHeader(self.HEADER)
        if not raw_header:
            return self._unauthorized(request)
        header = self.preprocess_header(raw_header)
        token_data = self.tokens.get(header, None)

        if token_data is None:
            return self._unauthorized(request)

        self._processToken(token_data, request).addCallback(  # type: ignore
            lambda _: request.finish()
        )
        return server.NOT_DONE_YET

    def preprocess_header(self, header: str) -> str:
        return header

    def processToken(self, token_data: Any, request: Request) -> int:  # noqa: ARG002
        """
        Process the token and write to request as needed.

        You are not allowed to call request.finish(), that will be handled
        by L{TokenResource._processToken}.

        You can return a L{defer.Deferred} which will be waited upon to
        finish the request.

        @param token_data: The object associated with the passed token.
        @param request: The request object associated to this request.
        """
        return OK

    @defer.inlineCallbacks
    def _processToken(self, token_data: Any, request: Request) -> Generator[defer.Deferred[int], None, None]:
        """
        Invoke L{TokenResource.processToken} and produce a 500 if it fails.

        @param token_data: The object associated with the passed token.
        @param request: The request object associated to this request.
        """
        try:
            res = yield defer.maybeDeferred(self.processToken, token_data, request)
            code: int = cast(int, res)
        except Exception:
            log.failure("Unknown error")

            code = INTERNAL_SERVER_ERROR
        request.setResponseCode(code)  # type: ignore
        defer.returnValue(code)

    def _unauthorized(self, request: Request) -> bytes:
        """
        Send a 401 Unauthorized response.

        @param request: The request object associated to this request.
        """
        request.setResponseCode(UNAUTHORIZED)  # type: ignore
        return self.unauthorizedPage().render(request)  # type: ignore

    def unauthorizedPage(self) -> resource.ErrorPage:
        """
        Page to render when there is no valid token.
        This makes use of L{TokenResource.unauthorizedMessage} by default.
        """
        return resource.ErrorPage(UNAUTHORIZED, "Unauthorized", self.unauthorizedMessage())

    def unauthorizedMessage(self) -> str:
        """
        Message to show when there is no valid token.
        """
        return f"Pass a valid token in the {self.HEADER} header."

    def getChild(self, name: str, request: Request) -> "TokenResource":  # noqa: ARG002
        """
        Use this child for everything but the explicitly overriden.

        @see: L{resource.Resource.getChild}
        """
        return self



from typing import Any, Dict, Generator, List, Optional, cast

import attr
import yaml
from twisted.internet import defer, reactor, task
from twisted.logger import Logger
from twisted.python.filepath import FilePath

from adlermanager.Config import ConfigClass
from adlermanager.IncidentManager import IncidentManager
from adlermanager.model import Alert, Severity, SiteConfig
from adlermanager.utils import TimestampFile, default_errback, noop_deferred


@attr.s
class SitesManager:
    global_config: ConfigClass = attr.ib()
    site_managers: Dict[str, "SiteManager"] = attr.ib(factory=dict)
    tokens: Dict[str, "SiteManager"] = attr.ib(factory=dict)
    log: Logger = attr.ib(factory=Logger)

    def __attrs_post_init__(self) -> None:
        def startup_message() -> None:
            self.log.info(
                "Starting server with this configuration:\n{self.global_config}",
                system=SitesManager.__name__,
            )

        # Inform people of current configuration when reactor starts
        _ = task.deferLater(reactor, 0, startup_message).addErrback(  # type: ignore
            default_errback
        )
        # Load data
        self.reload()

    def reload(self) -> "SitesManager":
        # Read sites
        read_sites = {
            site: self.site_managers[site].reload()
            if site in self.site_managers
            else SiteManager(global_config=self.global_config, path=self.sites_dir.child(site))
            for site in self.load_sites()
        }
        # Remove deleted sites
        for deleted_site in set(self.site_managers.keys()).difference(read_sites.keys()):
            del self.site_managers[deleted_site]
        # Apply update / add new sites
        self.site_managers.update(read_sites)
        # Re-read all sites
        self.tokens.clear()
        self.tokens.update({token: manager for manager in self.site_managers.values() for token in manager.tokens})
        return self

    @property
    def sites_dir(self) -> FilePath:
        return FilePath(self.global_config.data_dir).child("sites")

    def load_sites(self) -> Generator[str, None, None]:
        for site_dir in self.sites_dir.children():
            if not site_dir.isdir():
                continue
            yield cast(str, site_dir.basename())  # type: ignore

    def get_user_sites(self, username: bytes) -> Dict[str, "SiteManager"]:
        o: Dict[str, SiteManager] = {}
        try:
            u = username.decode("utf-8")
        except Exception:
            return o
        for k, sm in self.site_managers.items():
            if u in sm.ssh_users:
                o[k] = sm
        return o


@attr.s
class SiteManager:
    global_config: ConfigClass = attr.ib()
    path: FilePath = attr.ib()
    tokens: List[str] = attr.ib(factory=list)
    ssh_users: List[str] = attr.ib(factory=list)
    monitoring_is_down: bool = attr.ib(default=False)
    definition: Dict[str, Any] = attr.ib(factory=dict)
    title: str = attr.ib(default="")
    site_config: SiteConfig = attr.ib(factory=SiteConfig)
    service_managers: Dict[str, "ServiceManager"] = attr.ib(factory=dict)

    _timeout: defer.Deferred[None] = attr.ib(factory=noop_deferred)
    site_name: str = attr.ib(default="")

    @property
    def monitoring_down_seconds(self) -> float:
        return self.global_config.monitoring_down_minutes.total_seconds()

    log: Logger = attr.ib(factory=Logger)

    def __attrs_post_init__(self) -> None:
        self.reload()

    def reload(self) -> "SiteManager":
        self.load_definition()
        self.title = self.definition["title"]
        self.load_tokens()
        self.site_config = self.load_config()
        self.site_name = self.path.basename()
        self.last_updated = TimestampFile(self.path.child("last_updated.txt"))
        # Read services
        read_services: Dict[str, ServiceManager] = {
            s["label"]: self.service_managers[s["label"]].reload(s)
            if s["label"] in self.service_managers
            else ServiceManager(
                global_config=self.global_config,
                path=self.path.child(s["label"]),
                definition=s,
            )
            for s in cast(List[Dict[str, Any]], self.definition.get("services", {}))
        }
        # Remove deleted servicse
        for deleted_service in set(self.service_managers.keys()).difference(read_services.keys()):
            del self.service_managers[deleted_service]
        # Apply update / add new sites
        self.service_managers.update(read_services)

        # Add/reset monitoring timeout
        self._timeout.cancel()
        self._timeout = task.deferLater(
            reactor,  # type: ignore
            self.monitoring_down_seconds,
            self.monitoring_down,
        ).addErrback(default_errback)
        return self

    def monitoring_down(self) -> None:
        self.monitoring_is_down = True
        for manager in self.service_managers.values():
            manager.monitoring_down(self.last_updated.getStr())

    def load_definition(self) -> None:
        with self.path.child("site.yml").open("r") as f:
            self.definition = yaml.safe_load(f)
            self.ssh_users.clear()
            self.ssh_users.extend(u for u in self.definition.get("ssh_users", []))

    def load_tokens(self) -> None:
        tokens_file = self.path.child("tokens.txt")
        if tokens_file.exists():
            with open(tokens_file.path) as f:
                self.tokens = [line.strip() for line in f]
        if not self.tokens:
            self.log.warn(f"Site {self.title}: No tokens exist, your site will never update")

    @property
    def config_file(self) -> FilePath:
        return self.path.child("config.yaml")

    def load_config(self) -> SiteConfig:
        try:
            return SiteConfig.from_YAML(self.config_file.getContent())
        except Exception:
            return SiteConfig()

    def process_alerts(self, raw_alerts: List[Dict[str, Any]]) -> None:
        self.last_updated.now()

        self.monitoring_is_down = False
        self._timeout.cancel()
        self._timeout = task.deferLater(
            reactor,  # type: ignore
            self.monitoring_down_seconds,
            self.monitoring_down,
        ).addErrback(default_errback)

        # Filter alerts for this site
        alerts: List[Alert] = [
            Alert.import_alert(ra)
            for ra in raw_alerts
            if (
                ra.get("labels", {}).get("adlermanager", "") == self.site_name
                and ra.get("labels", {}).get("component", "")
                and ra.get("labels", {}).get("service", "")
            )
        ]

        heartbeats: List[Alert] = []
        filtered_alerts: List[Alert] = []
        for a in alerts:
            (heartbeats if a.labels.get("heartbeat") else filtered_alerts).append(a)

        timestamp = self.last_updated.getStr()
        for manager in self.service_managers.values():
            manager.process_heartbeats(heartbeats, timestamp)
            manager.process_alerts(filtered_alerts, timestamp)

    @property
    def status(self) -> Severity:
        if self.monitoring_is_down:
            return Severity.ERROR
        return max(
            (service.status for _, service in self.service_managers.items()),
            default=Severity.OK,
        )


@attr.s
class ServiceManager:
    global_config: ConfigClass = attr.ib()
    path: FilePath = attr.ib()
    definition: Dict[str, Any] = attr.ib()
    current_incident: Optional[IncidentManager] = attr.ib(default=None)
    component_labels: List[str] = attr.ib(factory=list)
    label: str = attr.ib(default="")

    def __attrs_post_init__(self) -> None:
        self.reload()

    def reload(self, definition: Optional[Dict[str, Any]] = None) -> "ServiceManager":
        if definition:
            # This should only happen when reloading
            self.definition.clear()
            self.definition.update(definition)
        self.label = self.definition["label"]
        self.component_labels.clear()
        self.component_labels.extend(
            [component["label"] for component in cast(List[Dict[str, Any]], self.definition.get("components", []))]
        )
        return self

    def monitoring_down(self, timestamp: str) -> None:
        if self.current_incident:
            self.current_incident.monitoring_down(timestamp)

    def process_heartbeats(self, heartbeats: List[Alert], timestamp: str) -> None:
        if self.current_incident:
            self.current_incident.process_heartbeats(heartbeats, timestamp)

    def process_alerts(self, alerts: List[Alert], timestamp: str) -> None:
        # Filter by service-affecting alerts
        alerts = [
            alert
            for alert in alerts
            if alert.labels.get("service") == self.label and alert.labels.get("component") in self.component_labels
        ]

        if alerts and not self.current_incident:
            # Something is up, open an incident
            self.current_incident = IncidentManager(
                global_config=self.global_config,
                path=self.path,
            )
            # Notify when incident is considered resolved
            _ = self.current_incident.expired.addCallback(self.resolve_incident)

        if self.current_incident:
            self.current_incident.process_alerts(alerts, timestamp)

    def resolve_incident(self, _: Any) -> None:
        self.current_incident = None

    @property
    def status(self) -> Severity:
        if self.current_incident:
            return max(
                (alert.status for alert in self.current_incident.active_alerts.values()),
                default=Severity.OK,
            )
        return Severity.OK

    # @property
    # def past_incidents(self):
    #     past = self.path.children().sort()
    #     if self.current_incident and self.current_incident.path in past:
    #         past.remove(self.current_incident.path)
    #     past.reverse()  # Beautiful sleepless code poetry
    #     # TODO: get limit from config?
    #     return past[:5]

    @property
    def components(self) -> List[Dict[str, Any]]:
        return [
            {
                "definition": component,
                "status": self.current_incident.component_status(component["label"])
                if self.current_incident
                else Severity.OK,
            }
            for component in self.definition.get("components", [])
        ]



import functools
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import attr
import yaml
from twisted.logger import Logger
from twisted.python.filepath import FilePath

from adlermanager.conch_helpers import SSHSimpleAvatar, SSHSimpleProtocol
from adlermanager.model import SiteConfig

if TYPE_CHECKING:
    from adlermanager.SitesManager import SiteManager, SitesManager

log = Logger()


class AdlerManagerSSHProtocol(SSHSimpleProtocol):
    sites_manager: "SitesManager"

    def __init__(self, user: SSHSimpleAvatar) -> None:
        """
        Create an instance of AdlerManagerSSHProtocol.
        """
        SSHSimpleProtocol.__init__(self, user)
        log.info("SSH login for {user}", user=user.username)

    def do_list_sites(self) -> None:
        """
        List all sites to which you have access
        """
        o: Dict[str, Any] = {}
        for k, sm in self.sites_manager.get_user_sites(self.user.username).items():
            o[k] = {
                "config": attr.asdict(sm.site_config),
                "status": {"value": sm.status.value, "message": str(sm.status)},
                "title": sm.title,
            }
        self.terminal_write(yaml.safe_dump(o, allow_unicode=True))
        self.terminal.nextLine()

    def _get_user_site_manager(self, site: bytes) -> Optional["SiteManager"]:
        try:
            s = site.decode("utf-8")
        except Exception:
            self.terminal_write("Error: Received invalid site name")
            self.terminal.nextLine()
            return None
        if s not in self.sites_manager.get_user_sites(self.user.username):
            self.terminal_write("Warning: requested unknown or unaccessible site")
            self.terminal.nextLine()
            return None
        return self.sites_manager.get_user_sites(self.user.username)[s]

    def do_get_site_config(self, site: bytes) -> None:
        """
        Get a site's configuration. Usage: get_site_config status.example.org
        """
        sm = self._get_user_site_manager(site)
        if sm is not None:
            self.terminal_write(sm.site_config.to_YAML())

    async def do_set_site_config(self, site: bytes) -> None:
        """
        Set a site's configuration.

        The second argument may be ommitted if running in a non-interactive
        fashion, in which case stdin will be used.

        Usage: set_site_config status.example.org \\
            [{"force_state": True, "message": "Hello world!\\n\\nThis is an example"}]
        """
        if self.interactive:
            self.terminal_write("Finish your YAML input with a line like this:")
            self.terminal.nextLine()
            self.terminal_write("---")
            self.terminal.nextLine()
        sm = self._get_user_site_manager(site)
        if sm is not None:
            data = await self.get_user_input(eom=b"---")
            if data is None or data == b"":
                msg = "No data was received"
                raise SyntaxError(msg)
            try:
                sc = SiteConfig.from_YAML(data)
            except Exception as ex:
                msg = "SiteConfig could not be created from data"
                raise SyntaxError(msg) from ex
            if self.interactive:
                self.terminal_write(sc.to_YAML())
                self.terminal_write("Does this look fine? [Y/n] ")
                ans = await self.get_user_input()
                if ans.decode("utf-8").strip().upper() not in ["Y", ""]:
                    self.terminal_write("Aborting")
                    self.terminal.nextLine()
                    return
            # Actually do something
            sm.site_config = sc
            sm.config_file.setContent(sc.to_YAML().encode("utf-8"))
            sm.config_file.chmod(0o640)
            log.info("User {user} changed {site} config", user=self.user.username, site=site)
            if self.interactive:
                self.terminal_write("Persisted SiteConfiguration")
                self.terminal.nextLine()

    def motd(self) -> Union[str, bytes]:
        return _read_motd(FilePath(self.sites_manager.global_config.data_dir).child("motd.txt"))


@functools.lru_cache  # we don't need to re-read every time
def _read_motd(custom_motd: FilePath) -> Union[str, bytes]:
    if custom_motd.isfile():
        return custom_motd.getContent()
    # Default motd
    return FilePath(__file__).parent().child("motd.txt").getContent()



#!/usr/bin/env python3
"""
Run in foreground using:
- a generic service manager: python -m adlermanager
- twistd: twistd -ny src/adlermanager/__main__.py
"""

from typing import Iterable, cast

from twisted.application import service, strports
from twisted.python.filepath import FilePath
from twisted.web import server

from adlermanager.Config import Config
from adlermanager.SitesManager import SitesManager
from adlermanager.WebRoot import web_root

if not FilePath(Config.data_dir).isdir():
    FilePath(Config.data_dir).createDirectory()

application = service.Application("AdlerManager")
serv_collection = service.IServiceCollection(application)

# TokenResource
sites_manager = SitesManager(global_config=Config)

resource = web_root(sites_manager)
site = server.Site(resource)
i = strports.service(Config.web_endpoint, site)  # type: ignore
i.setServiceParent(serv_collection)  # type: ignore

if Config.ssh_enabled:
    # Set up SSH config service
    from adlermanager.AdlerManagerSSHProtocol import AdlerManagerSSHProtocol
    from adlermanager.conch_helpers import conch_helper

    # TODO: Make this more decent
    AdlerManagerSSHProtocol.sites_manager = sites_manager
    i = conch_helper(
        Config.ssh_endpoint,
        proto=AdlerManagerSSHProtocol,  # type: ignore
        keyDir=Config.ssh_keys_dir,
        keySize=Config.ssh_key_size,
    )
    i.setServiceParent(serv_collection)  # type: ignore


def run() -> None:
    import sys

    from twisted import logger

    # Setup global logging to stdout
    logger.globalLogPublisher.addObserver(cast(logger.ILogObserver, logger.textFileLogObserver(sys.stdout)))

    # Run without twistd as documented upstream:
    # https://docs.twisted.org/en/stable/core/howto/basics.html#twistd
    from twisted.internet import reactor as _reactor  # type: ignore
    from twisted.internet.base import ReactorBase

    reactor = cast(ReactorBase, _reactor)
    for srv in cast(Iterable[service.IService], serv_collection):
        srv.startService()
        reactor.addSystemEventTrigger("before", "shutdown", srv.stopService)

    # Finally, start the reactor
    reactor.run()


if __name__ == "__main__":
    run()



from io import BytesIO
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from twisted.application import strports
from twisted.application.internet import StreamServerEndpointService
from twisted.conch import avatar, manhole_ssh, recvline
from twisted.conch import interfaces as conchinterfaces
from twisted.conch.checkers import (
    IAuthorizedKeysDB,
    SSHPublicKeyChecker,
    readAuthorizedKeyFile,
)
from twisted.conch.insults import insults
from twisted.conch.manhole_tap import chainedProtocolFactory
from twisted.conch.ssh import keys, session
from twisted.cred import portal
from twisted.internet import defer
from twisted.internet.error import ProcessTerminated
from twisted.python import failure, filepath
from zope.interface import implementer


class SSHSimpleProtocol(recvline.HistoricRecvLine):
    terminal: insults.ServerProtocol
    keyHandlers: Dict[bytes, Callable[[], None]]
    mode: str
    lineBuffer: List[bytes]
    _unicodeCharBuffer: List[bytes]
    _unicodeCharLength: int

    _command_lineReceived: Optional[Callable[[bytes], None]] = None
    _command_inputEnded: Optional[Callable[[bytes], None]] = None

    def __init__(self, user: "SSHSimpleAvatar") -> None:
        recvline.HistoricRecvLine.__init__(self)
        self.user = user
        self.ps = (b">>> ", b"... ")
        self._unicodeCharBuffer = []
        self._unicodeCharLength = 1

    def terminal_write(self, msg: Union[bytes, str]) -> None:
        """
        Write data to the remote terminal.

        We use this method to work around awkward typing in t.c.insults.
        """
        return self.terminal.write(msg)  # type: ignore

    def initializeScreen(self) -> None:
        """
        Initialise terminal in InsertMode, without clearing the screen.

        This overrides the definition in twisted.conch.recvline.RecvLine:
        https://github.com/twisted/twisted/blob/7697871b4d89c78c8764c6be42372fc68299714e/src/twisted/conch/recvline.py#L385
        """
        # This seems to produce minor issues on the client's terminal
        # self.setInsertMode()
        # But conch does assume .mode is set early
        self.mode = "insert"
        if self.interactive:
            self.terminal_write(self.motd())
            self.terminal.nextLine()
            self.showPrompt()

    def keystrokeReceived(self, keyID: bytes, modifier: Any) -> None:  # noqa: ARG002
        """
        Handle a received keystroke.

        The default implementation of this was keeping us from using anything
        not listed in string.printable, which is a rather limited subset of
        symbols.
        """
        # For those of us that forget periodically how unicode works:
        # https://docs.python.org/3/library/codecs.html#encodings-and-unicode
        # The table with the encoding is particularly useful :-)
        unicode_mark = (keyID[0] >> 7) & 1 == 1  # 0x1XXXYYYY
        m = self.keyHandlers.get(keyID)
        if m is not None:
            m()
        elif (
            not unicode_mark  # This is not a unicode byte
            and not self._unicodeCharBuffer  # Check we had nothing in buffer
            and keyID in self._printableChars  # Printable ascii
        ):
            self.characterReceived(keyID, False)  # type: ignore
        elif unicode_mark:  # Handle unicode
            leading_ones = 1  # 0x10XXYYYY .. 0x11110YYY
            for i in (6, 5, 4, 3):
                if keyID[0] >> i & 1 == 1:
                    leading_ones += 1
                else:
                    break
            if (  # Catch misbehaved input
                leading_ones > 4  # This is too many ones
                or (leading_ones < 2 and len(self._unicodeCharBuffer) < 1)  # This is a wrong starting byte
                or (
                    leading_ones > 1 and len(self._unicodeCharBuffer) > 0
                )  # This  ends a previous multi-byte char too soon
                # TODO: should we keep this starting byte in the buffer?
            ):
                self._unicodeCharBuffer.append(keyID)
                self._log.warn(
                    "Received unhandled unicode sequence: {bs!r}",
                    bs=b"".join(self._unicodeCharBuffer),
                )
                self._unicodeCharBuffer.clear()
            else:
                if leading_ones > 1:  # First byte of char
                    self._unicodeCharBuffer.clear()
                    # Save how many bytes we need
                    self._unicodeCharLength = leading_ones

                self._unicodeCharBuffer.append(keyID)
                if len(self._unicodeCharBuffer) == self._unicodeCharLength:
                    # Got enough bytes, send the char
                    self.characterReceived(  # type: ignore
                        b"".join(self._unicodeCharBuffer), False
                    )
                    self._unicodeCharBuffer.clear()
        else:
            self._log.warn("Received unhandled keyID: {keyID!r}", keyID=keyID)
            if self._unicodeCharBuffer:
                self._log.warn(
                    "    after unicode sequence: {bs!r}",
                    bs=b"".join(self._unicodeCharBuffer),
                )
            self._unicodeCharBuffer.clear()

        if not unicode_mark:
            self._unicodeCharBuffer.clear()

    def connectionMade(self) -> None:
        recvline.HistoricRecvLine.connectionMade(self)
        if not self.interactive:
            return
        # CTRL_D
        self.keyHandlers[b"\x04"] = self.handle_EOF
        # CTRL_BACKSLASH
        self.keyHandlers[b"\x1c"] = self.handle_QUIT

    def handle_EOF(self) -> None:
        if self._command_inputEnded is not None:
            self._command_inputEnded(b"".join(self.lineBuffer))
        if self.lineBuffer:  # type: ignore
            self.terminal_write(b"\a")
        else:
            self.exitWithCode(0)

    def handle_QUIT(self) -> None:
        self.terminal.loseConnection()

    @property
    def interactive(self) -> bool:
        """
        Whether or not we are running an interactive shell.

        This is currently done by determining if a pty has been requested.
        """
        return self.user.ptyAllocated

    @property
    def prompt(self) -> bytes:
        return self.ps[self.pn]

    def showPrompt(self) -> None:
        if self.interactive:
            self.terminal_write(self.prompt)

    def _getCommand(self, cmd: bytes) -> Optional[Callable[..., Union[None, Coroutine[Any, Any, None]]]]:
        """
        Get the method that would be run by 'cmd' if appliable.

        The convention is that a cmd command translates to a do_cmd method.
        """
        try:
            cmd_str = cmd.decode("utf-8")
            return getattr(self, "do_" + cmd_str, None)
        except Exception:
            return None

    async def runCommand(self, cmd: bytes, *args: bytes) -> None:
        """
        Run the requested command or print 'No such command'.
        """
        func = self._getCommand(cmd)
        if func:
            try:
                res = func(*args)
                if res is not None:
                    await defer.ensureDeferred(res)
                if not self.interactive:
                    self.exitWithCode(0)
            except Exception as ex:
                self.terminal_write(f"Error: {ex}")
                self.terminal.nextLine()
                if not self.interactive:
                    self.exitWithCode(2)
        else:
            self.terminal_write(b"No such command: " + cmd)
            self.terminal.nextLine()
            if not self.interactive:
                self.exitWithCode(1)

    async def runLine(self, line: bytes) -> None:
        line_parts = line.strip().split()
        if line_parts:
            cmd, *args = line_parts
            return await self.runCommand(cmd, *args)

    async def _lineReceived(self, line: bytes) -> None:
        await self.runLine(line)
        self.showPrompt()

    def lineReceived(self, line: bytes) -> None:
        if self._command_lineReceived is None:
            _ = defer.ensureDeferred(self._lineReceived(line))
        else:
            self._command_lineReceived(line)

    def get_user_input(self, eom: bytes = b"") -> defer.Deferred[bytes]:
        d = defer.Deferred[bytes]()
        data = BytesIO()

        def _ui_lineReceived(line: bytes) -> None:
            if eom and line != eom:
                data.write(line)
                data.write(b"\n")
            else:
                _ui_inputEnded(line)

        def _ui_inputEnded(line: bytes) -> None:
            if line and line != eom:
                data.write(line)
                data.write(b"\n")
            self._command_lineReceived, self._command_inputEnded = None, None
            d.callback(data.getvalue())

        self._command_lineReceived, self._command_inputEnded = (
            _ui_lineReceived,
            _ui_inputEnded,
        )
        return d

    def exitWithCode(self, code: int) -> None:
        cast(session.SSHSessionProcessProtocol, self.terminal.transport).processEnded(
            failure.Failure(ProcessTerminated(code, None, None))
        )

    def do_help(self, cmd: bytes = b"") -> None:
        """
        Get help on a command. Usage: help [command]
        """
        if cmd:
            func = self._getCommand(cmd)
            if func:
                self.terminal_write(getattr(func, "__doc__", b"No help for '" + cmd + b"'"))
            else:
                self.terminal_write(b"No such command: " + cmd)
        else:
            self.terminal_write("Available commands:")
            self.terminal.nextLine()
            self.terminal.nextLine()
            for c in (attr[3:] for attr in dir(self) if attr.startswith("do_")):
                self.terminal_write(c)
                self.terminal.nextLine()
        self.terminal.nextLine()

    def do_whoami(self) -> None:
        """
        Prints your username. Usage: whoami
        """
        self.terminal_write(self.user.username)
        self.terminal.nextLine()

    def do_clear(self) -> None:
        """
        Clears the screen. Usage: clear
        """
        self.terminal.reset()

    def do_exit(self) -> None:
        """
        Exit session. Usage: exit
        """
        self.exitWithCode(0)

    def motd(self) -> Union[str, bytes]:
        return ""


@implementer(conchinterfaces.ISession)
class SSHSimpleAvatar(avatar.ConchUser):
    serverProtocol: Optional[insults.ServerProtocol] = None
    ptyAllocated: bool = False
    """
    Whether or not a pty was ever requested.
    """

    def __init__(self, username: bytes, proto: SSHSimpleProtocol):
        avatar.ConchUser.__init__(self)

        self.username = username
        self.proto = proto
        self.channelLookup.update({b"session": session.SSHSession})  # type: ignore

    def openShell(self, protocol: session.SSHSessionProcessProtocol) -> None:
        self.serverProtocol = insults.ServerProtocol(self.proto, self)
        self.serverProtocol.makeConnection(protocol)  # type: ignore
        protocol.makeConnection(  # type: ignore
            session.wrapProtocol(self.serverProtocol)  # type: ignore
        )

    def getPty(self, terminal, windowSize, attrs) -> None:  # type: ignore  # noqa: ARG002
        # Save the pty request for use from the protocol
        self.ptyAllocated = True

    def execCommand(self, protocol: session.SSHSessionProcessProtocol, line: bytes) -> None:
        self.serverProtocol = insults.ServerProtocol(self.proto, self)
        self.serverProtocol.makeConnection(protocol)  # type: ignore
        protocol.makeConnection(  # type: ignore
            session.wrapProtocol(self.serverProtocol)  # type: ignore
        )
        _ = defer.ensureDeferred(
            cast(
                SSHSimpleProtocol,
                self.serverProtocol.terminalProtocol,  # type: ignore
            ).runLine(line)
        )

    def windowChanged(self, dimensions: Tuple[int, int, int, int]):  # type: ignore
        """This gets triggered when user changes terminal dimensions.

        Args:
            dimensions (Tuple[int, int, int, int]): Probably height, width and "0, 0"
        """

    def eofReceived(self) -> None:
        """
        Called when the other side has indicated no more data will be sent.
        """
        if self.serverProtocol is not None:
            ssh_sp = cast(
                SSHSimpleProtocol,
                self.serverProtocol.terminalProtocol,  # type: ignore
            )
            ssh_sp.handle_EOF()

    def closed(self) -> None:
        """
        Called when the session is closed.
        """


@implementer(portal.IRealm)
class SSHSimpleRealm:
    """
    SSH simple realm that uses a given protocol for any valid users.

    You may want to customise the way that protocol works, realm and avatar
    are not necessarily of interest here.

    @ivar proto: The passed protocol class that will be used for avatar.
    """

    def __init__(self, proto: SSHSimpleProtocol):
        """
        Initialise a new L{SSHSimpleRealm} with proto as a protocol class.

        @param proto: a protocol class that will be used for any avatars.
        """
        self.proto = proto

    def requestAvatar(
        self,
        avatarId: Union[bytes, Tuple[()]],
        mind: object,  # noqa: ARG002
        *interfaces: portal._InterfaceItself,  # type: ignore
    ) -> Union[
        defer.Deferred[portal._requestResult], portal._requestResult  # type: ignore
    ]:
        """
        Return a L{SSHSimpleAvatar} that uses ``self.proto`` as protocol.

        @see: L{portal.IRealm}
        """
        if conchinterfaces.IConchUser in interfaces:
            avatar = SSHSimpleAvatar(cast(bytes, avatarId), self.proto)
            return interfaces[0], avatar, lambda: None
        else:
            msg = "No supported interfaces found."
            raise Exception(msg)


@implementer(IAuthorizedKeysDB)
class SSHKeyDirectory:
    """
    Provides SSH public keys based on a simple directory structure.

    For a user ``USER`` following files are returned if they exist:
      - ``$USER/*.key``
      - ``$USER.key``
    These paths are relative to L{SSHKeyDirectory.baseDir}

    @ivar baseDir: the base directory for key lookup.
    """

    def __init__(
        self,
        baseDir: filepath.FilePath[str],
        parseKey: Any = keys.Key.fromString,  # type: ignore
    ) -> None:
        """
        Initialises a new L{SSHKeyDirectory}.

        @param base_dir: the base directory for key lookup.
        @param parseKey: L{callable}
        """
        self.baseDir = baseDir
        self.parseKey = parseKey

    def getAuthorizedKeys(self, username: bytes) -> Iterator[keys.Key]:
        keyFile = self.baseDir.child(username + b".key")
        keyDir = self.baseDir.child(username)

        if keyFile.isfile():
            for key in readAuthorizedKeyFile(keyFile.open(), self.parseKey):
                yield key

        if keyDir.isdir():
            for f in keyDir.globChildren("*.key"):
                for key in readAuthorizedKeyFile(f.open(), self.parseKey):
                    yield key


def conch_helper(
    endpoint: str,
    proto: Optional[SSHSimpleProtocol] = None,
    namespace: Optional[Dict[str, str]] = None,
    keyDir: Optional[str] = None,
    keySize: int = 4096,
) -> StreamServerEndpointService:
    """
    Return a L{SSHKeyDirectory} based SSH service with the given parameters.

    Authorized keys are read as per L{SSHKeyDirectory} with ``baseDir`` being
    ``keyDir/users``.

    @param endpoint: endpoint for the SSH service
    @param namespace: the manhole namespace
    @param keyDir: directory that holds server/server.key file and
        users directory, which is used as ``baseDir`` in L{SSHKeyDirectory}
    @see: L{SSHKeyDirectory}
    """
    namespace = {} if namespace is None else namespace
    if keyDir is None:
        from twisted.python._appdirs import getDataDirectory

        keyDir = getDataDirectory()

    ssh_keys_dir = filepath.FilePath(keyDir)
    ssh_keys_dir.child("server").makedirs(True)
    ssh_keys_dir.child("users").makedirs(True)

    checker = SSHPublicKeyChecker(
        SSHKeyDirectory(ssh_keys_dir.child("users"))  # type: ignore
    )

    if proto is None:
        sshRealm = manhole_ssh.TerminalRealm()
        sshRealm.chainedProtocolFactory = chainedProtocolFactory(  # type: ignore
            namespace
        )
    else:
        sshRealm = SSHSimpleRealm(proto)  # type: ignore
    sshPortal = portal.Portal(sshRealm, [checker])  # type: ignore

    sshKeyPath = ssh_keys_dir.child("server").child("server.key")
    sshKey = keys._getPersistentRSAKey(sshKeyPath, keySize)  # type: ignore

    sshFactory = manhole_ssh.ConchFactory(sshPortal)
    sshFactory.publicKeys[b"ssh-rsa"] = sshKey  # type: ignore
    sshFactory.privateKeys[b"ssh-rsa"] = sshKey  # type: ignore

    sshService = strports.service(endpoint, sshFactory)  # type: ignore
    return sshService
